# -*- coding: utf-8 -*-

import json
import os.path
import requests
import sys

from datetime import datetime
from easymaker.common import constants
from easymaker.common import exceptions


class ObjectStorage:
    CHUNK_SIZE = 100 * 1024  # 100 KB
    FILE_SIZE_LIMIT = 5 * 1024 * 1024 * 1024  # 5 GB

    # TODO. username,password 없이 API 서버에서 iaas 토큰 받아서 다운로드하는 형태로 변경 필요
    def __init__(self, tenant_id, username=None, password=None, region=constants.DEFAULT_REGION):
        self.tenant_id = tenant_id
        self.username = username
        self.password = password
        self.token_expires = None
        self.region = region
        self._get_token()

    def _get_token(self):
        if self.token_expires is not None:
            time_diff = self.token_expires - datetime.utcnow()
            if time_diff.total_seconds() > 600:
                return
        token_url = constants.OBJECT_STORAGE_AUTH_URL + '/tokens'
        req_header = {'Content-Type': 'application/json'}
        req_body = {
            'auth': {
                'tenantId': self.tenant_id,
                'passwordCredentials': {
                    'username': self.username,
                    'password': self.password
                }
            }
        }

        response = requests.post(token_url, headers=req_header, json=req_body).json()
        self._get_storage_url(response)
        self.token = response['access']['token']
        self.token_id = self.token['id']
        self.token_expires = datetime.strptime(self.token['expires'], '%Y-%m-%dT%H:%M:%SZ')

    def _get_storage_url(self, response):
        for serviceCatalog in response['access']['serviceCatalog']:
            if serviceCatalog['type'] == 'object-store':
                for endpoint in serviceCatalog['endpoints']:
                    if endpoint['region'].casefold() == self.region.casefold():
                        self.storage_url = endpoint['publicURL']

    def _get_url(self, container, object=None):
        if object is None:
            return '/'.join([self.storage_url, container])
        return '/'.join([self.storage_url, container, object])

    def _get_request_header(self):
        return {'X-Auth-Token': self.token_id}

    def upload(self, container, object_dir_path, upload_path):
        """
        Args:
            container : obs container name
            object_dir_path : obs object path (directory)
            upload_path : upload local path (file or directory)
        """
        self._get_token()

        if os.path.isfile(upload_path):
            object_file_path = os.path.join(object_dir_path, os.path.basename(upload_path))
            self._upload_file(container, object_file_path, upload_path)

        if os.path.isdir(upload_path):
            file_path_list = []
            for (root, dirs, files) in os.walk(upload_path):
                for file in files:
                    file_path_list.append(os.path.join(root, file))

            for upload_file_path in file_path_list:
                object_file_path = os.path.join(object_dir_path,
                                                os.path.relpath(upload_file_path,
                                                                os.path.abspath(upload_path)))
                self._upload_file(container, object_file_path, upload_file_path)

    def _upload_file(self, container, object_file_path, upload_file_path):
        """
        Upload files under 5G
        Args:
            container : obs container name
            object_file_path : obs object path (file)
            upload_file_path : upload local path (file)
        """
        if os.path.getsize(upload_file_path) >= self.FILE_SIZE_LIMIT:
            return _upload_large_file(container, object_file_path, upload_file_path)

        self._get_token()
        req_url = self._get_url(container, object_file_path)
        req_header = self._get_request_header()

        with open(upload_file_path, 'rb') as f:
            return requests.put(req_url, headers=req_header, data=f.read())

    def _upload_large_file(self, container, object_file_path, upload_file_path):
        """
        Objects with a capacity exceeding 5 GB are uploaded in segments of 5 GB or less.
        """
        self._get_token()
        url = self._get_url(container, object_file_path)
        req_header = self._get_request_header()

        with open(upload_file_path, 'rb') as f:
            chunk_index = 1
            chunk_size = self.CHUNK_SIZE
            total_bytes_read = 0
            obj_size = os.path.getsize(upload_file_path)

            while total_bytes_read < obj_size:
                remained_bytes = obj_size - total_bytes_read
                if remained_bytes < chunk_size:
                    chunk_size = remained_bytes

                req_url = '%s/%03d' % (url, chunk_index)
                requests.put(
                    req_url, headers=req_header, data=f.read(chunk_size))
                total_bytes_read += chunk_size
                f.seek(total_bytes_read)
                chunk_index += 1

        # create manifest
        req_url = self._get_url(container, object_file_path)
        req_header = self._get_request_header()
        req_header['X-Object-Manifest'] = '/'.join([container, object_file_path])
        return requests.put(req_url, headers=req_header)

    def _get_object_list(self, container, object_path):
        req_url = self._get_url(container)
        req_header = self._get_request_header()
        response = requests.get(req_url, headers=req_header, params={'prefix': object_path})
        return response.text.split('\n')

    def download(self, container, object_prefix, download_dir_path):
        """
        Args:
            container : obs container name
            object_prefix : obs object path (file or directory)
            download_dir_path : download local path (directory)
        """
        self._get_token()
        isDirectoryObject = False
        file_object_list = []

        objectList = self._get_object_list(container, object_prefix)
        for obj in objectList:

            if (object_prefix.endswith('/') == False) and (obj == object_prefix):  # target object is file
                download_file_path = os.path.join(download_dir_path, os.path.basename(object_prefix))
                # object : depth1/file1
                # download_file_path => download_dir_path + /file1
                return self._download_file(container, object_prefix, download_file_path)

            if object_prefix.endswith('/') == False:
                object_prefix = ''.join([object_prefix, '/'])

            if obj.startswith(object_prefix):
                isDirectoryObject = True
                if not obj.endswith('/'):
                    file_object_list.append(obj)

        if isDirectoryObject:
            for file_object in file_object_list:
                download_file_path = os.path.join(download_dir_path, os.path.relpath(file_object, object_prefix))
                # object : deps1/deps2, file_object : deps1/deps2/deps3/file1
                # download_file_path => download_dir_path + /deps3/file1
                self._download_file(container, file_object, download_file_path)

    def _download_file(self, container, object, download_file_path):
        """
        Args:
            container : obs container name
            object : obs object path (file)
            download_file_path : download local path (file)
        """
        req_url = self._get_url(container, object)
        req_header = self._get_request_header()
        response = requests.get(req_url, headers=req_header)

        if response.status_code != 200:
            raise exceptions.EasyMakerError(f'object storage donwload fail {response.json()}')

        os.makedirs(os.path.dirname(download_file_path), exist_ok=True)
        with open(download_file_path, 'wb') as f:
            f.write(response.content)


def obs_uri_parser(obs_uri):
    uri_split = obs_uri.split("://")

    if len(uri_split) != 2 or uri_split[0] != 'obs':
        raise exceptions.EasyMakerError(f'object storage uri parse fail. Invalid uri {obs_uri}')

    uri_list = uri_split[1].split('/')

    if len(uri_list) < 3 or not uri_list[0] or not uri_list[1]:
        raise exceptions.EasyMakerError(f'object storage uri parse fail. Invalid uri {obs_uri}')

    print (f'uri_list[2] {uri_list[2]}, {uri_list[2].startswith("/")}')

    return uri_list[0], uri_list[1], '/'.join(uri_list[2:])

def download(obs_uri, download_dir_path):  # obs_uri : 'obs://tenant_id/container_name/path'
    tenant_id, container_name, object_prefix = obs_uri_parser(obs_uri)

    object_storage = ObjectStorage(tenant_id)
    object_storage.download(container_name, object_prefix, download_dir_path)