from typing import Optional
import easymaker
from easymaker.training import model
from easymaker.storage import objectstorage
from easymaker.image import docker_builder
from easymaker.common import constants
from easymaker.common import utils


class TrainingJob:
    _framework_name = None
    _framework_version = None

    def __init__(
            self,
            container_registry=None,
            container_username=None,
            container_password=None,
            base_image=None,
            training_image=None,
            docker_file=None,
            contents_dir='.',
            work_dir=constants.DEFAULT_WORKDIR,
            docker_command=None,
    ):
        self.container_registry = container_registry
        self.container_username = container_username
        self.container_password = container_password

        self._docker_builder = docker_builder.DockerBuilder(self.container_registry,
                                                            self.container_username,
                                                            self.container_password)
        self.base_image = base_image
        self.training_image = training_image

        # training image build and push
        self._docker_builder.build(build_image=training_image,
                                   dockerfile_path=docker_file,
                                   base_image=base_image,
                                   contents_dir=contents_dir,
                                   work_dir=work_dir,
                                   docker_command=docker_command,
                                   )
        self._docker_builder.push(training_image)

    def run(self,
            training_name=None,
            training_description=None,
            model_upload_uri=None,
            checkpoint_upload_uri=None,
            channel_list=None,
            tag_list=None,
            training_type_code=None,
            train_flavor_name=None,
            train_instance_count=None,
            boot_storage_size=None,
            dataStorage_type=None,
            dataStorage_size=None,
            wait=False,
            ) -> Optional[model.Model]:

        # run training
        response = easymaker.api_sender.run_training(
            {
                # training info
                'trainingName': training_name,
                'trainingDescription': training_description,
                'trainingTimeLimit': training_name,
                'modelUploadUri': model_upload_uri,
                'checkpointUploadUri': checkpoint_upload_uri,
                'channelList': channel_list,
                'tagList': tag_list,  # Key-Value 형태, 최대 5개
                'trainingTypeCode': training_type_code,
                # registry
                'baseImageUri': self.base_image,
                'trainingImageUri': self.training_image,
                'containerRegistryUri': self.container_registry,
                'containerUsername': self.container_username,
                'containerPassword': self.container_password,
                # IaaS
                'flavorId': utils.from_flavor_name_to_flavor_id(train_flavor_name),
                'instanceCount': train_instance_count,
                'bootStorageSize': boot_storage_size,
                'dataStorageType': dataStorage_type,
                'dataStorageSize': dataStorage_size,
            }
        )

        if wait:
            # todo training 종료 확인 loop
            return model.Model()
        else:
            return response
